-- require "./AI/Const.lua" 

 
--------------------------------------------
-- List utility 
--------------------------------------------
List = {}

	
function List.new () 
	return { first = 0, last = -1}
end


function List.pushleft (list, value) 
	local first = list.first-1
	list.first  = first
	list[first] = value;
end

 
function List.pushright (list, value) 
	local last = list.last + 1
	list.last = last
	list[last] = value
end


function List.popleft (list) 
	local first = list.first
	if first > list.last then
		return nil
	end
	local value = list[first]
	list[first] = nil         -- to allow garbage collection
	list.first = first+1
	return value
end


function List.popright (list) 
	local last = list.last
	if list.first > last then
		return nil
	end
	local value = list[last]
	list[last] = nil
	list.last = last-1
	return value
end


function List.clear (list) 
	for i,v in ipairs(list) do
		list[i] = nil
	end
--[[
	if List.size(list) == 0 then
		return
	end
	local first = list.first
	local last  = list.last
	for i=first, last do
		list[i] = nil
	end
--]]
	list.first = 0
	list.last = -1
end


function List.size (list) 
	local size = list.last - list.first + 1
	return size
end

-------------------------------------------------

function List.getn(t)
	local l = 0
	for i,v in pairs(t) do
		l = l + 1
	end

	return l
end


-------------------------------------------------  
function	GetDistance (x1,y1,x2,y2)
--	return math.floor(math.sqrt((x1-x2)^2+(y1-y2)^2))
	local xd,yd = math.abs(x2-x1),math.abs(y2-y1)

	if xd > yd then
		return xd
	else
		return yd
	end
end

 
function	GetDistance2 (id1, id2) 
	local x1,y1 = GetV (V_POSITION,id1)
	local x2,y2 = GetV (V_POSITION,id2)
	if (x1 == -1 or x2 == -1) then
		return -1
	end
	return GetDistance (x1,y1,x2,y2)
end

 
-------------------------------------------------  

-- Function to get the angle from one point to another
function GetAngle(x1,y1,x2,y2)
--	TraceAI ("x1,y1="..x1..","..y1.." x2,y2="..x2..","..y2)

	-- Get the radius
--	local radius = GetDistance(x1,y1,x2,y2)
	local radius = math.sqrt((x1-x2)^2+(y1-y2)^2)

	-- Check if the radius is 0
	if radius == 0 then
		-- Return an angle of -1
		return -1,0
	end

	-- Get the changes
	local xdelta = x2 - x1
	local ydelta = y2 - y1

	TraceAI (string.format ("radius=%f, xdelta,ydelta=%d,%d asinx=%f acosy=%f",radius,xdelta,ydelta,math.deg(math.asin(xdelta / radius)),math.deg(math.acos(ydelta / radius))))

	-- Get the angle from Y
	local angle = math.deg(math.acos(ydelta / radius))

	-- Get the angle from X
	if math.deg(math.asin(xdelta / radius)) < 0 then
		angle = 360 - angle
	end

	-- Return the angle and the radius
	return angle,radius
end


-- Function to plot a point on a circle centered around a position
function PlotCircle(x,y,angle,radius)
	-- Get the angle in radians
	local rads = math.rad(angle)

	-- Get the xdelta and ydeltas
	local xdelta = (radius * math.sin(rads))
	local ydelta = (radius * math.cos(rads))

	-- Return the locations
	return x + xdelta, y + ydelta
end


-- Function to detect direction from x1,y1 to x2,y2 eAthena-style
function CalcDirection (x1,y1,x2,y2)
	local dx,dy = x1-x2,y1-y2
	local dir

	if (dx == 0 and dy == 0) then
		-- both are standing on the same spot
		dir = 0
	elseif (dx >= 0 and dy >=0) then	-- upper-right
		if (dx*2 <= dy) then
			dir = 0		-- up
		elseif (dx > dy*2) then
			dir = 6		-- right
		else
			dir = 7		-- up-right
		end
	elseif (dx >= 0 and dy <= 0) then	-- lower-right
		if (dx*2 <= -dy) then
			dir = 4		-- down
		elseif (dx > -dy*2) then
			dir = 6		-- right
		else
			dir = 5		-- down-right
		end
	elseif (dx <= 0 and dy <= 0) then	-- lower-left
		if (dx*2 >= dy) then
			dir = 4		-- down
		elseif (dx < dy*2) then
			dir = 2		-- left
		else
			dir = 3		-- down-left
		end
	else					-- upper-left
		if (-dx*2 <= dy) then
			dir = 0		-- up
		elseif (-dx > dy*2) then
			dir = 2		-- left
		else
			dir = 1		-- up-left
		end
	end

	return dir	
end

-------------------------------------------------  

function	GetOwnerPosition (id) 
	return GetV (V_POSITION,GetV(V_OWNER,id))
end

 
function	GetDistanceFromOwner (id) 
	local x1, y1 = GetOwnerPosition (id)
	local x2, y2 = GetV (V_POSITION,id)
	if (x1 == -1 or x2 == -1) then
		return -1
	end
	return GetDistance (x1,y1,x2,y2)
end

 
function	GetDistanceFromMaster (myid) 
------------------------------------------
-- }X^[Ƃ̋vZ
------------------------------------------
	local x1, y1 = GetV (V_POSITION,MyMaster)
	local x2, y2 = GetV (V_POSITION,myid)
	if (x1 == -1 or x2 == -1) then
		return -1
	end
	return GetDistance (x1,y1,x2,y2)
end


-------------------------------------------------  
 
function	IsOutOfSight (id1,id2) 
	local x1,y1 = GetV (V_POSITION,id1)
	local x2,y2 = GetV (V_POSITION,id2)
	if (x1 == -1 or x2 == -1) then
		return true
	end
	local d = GetDistance (x1,y1,x2,y2)
	if d > 14 then
		return true
	else
		return false
	end
end


function	IsInAttackSight (id1,id2) 
	local x1,y1 = GetV (V_POSITION,id1)
	local x2,y2 = GetV (V_POSITION,id2)
	if (x1 == -1 or x2 == -1) then
		return false
	end

	local d	= GetDistance (x1,y1,x2,y2)
	local a = GetV (V_ATTACKRANGE,id1)

	if a >= d then
		return true;
	else
		return false;
	end
end

 
function	IsInSkillSight (id1,id2,skill) 
	local x1,y1 = GetV (V_POSITION,id1)
	local x2,y2 = GetV (V_POSITION,id2)
	if (x1 == -1 or x2 == -1) then
		return false
	end
	local d	= GetDistance (x1,y1,x2,y2)
	local a = GetV (V_SKILLATTACKRANGE,id1,skill)

	if (skill == SKILL_MAGNUM) then
		a = 2
	end

	if a >= d then
		return true
	else
		return false
	end
end


-- IsȂ񂿂֐ IuWFNgޔʌn --------------- 
	
function GetType (id)
	local type = GetV(V_HOMUNTYPE,id)					-- id'_^_C_v''_''_B

	if (type == nil) then	--     
		type = MobID[id]
	end

	if (type ~= nil) then
		return type
	end
	return -1
end


function IsPlayer (id) 
------------------------------------------
-- ΏۂvC[f (true:vC[ false:̑)
-- zid < 100000 Ƃ񍐂BŃvC[ƃz𔻕ʁB
------------------------------------------
	if (IsMonster (id) ~= 1) then
		if (id < 100000 or id == -1) then
			return false
		else
			local type = GetV (V_HOMUNTYPE,id)

			if (type == nil) then	--     
				type = MobID[id]
			end

			if ((type == nil) or (0 <= type and type <= 25) or
				(161 <= type and type <= 181) or
				(4001 <= type and type <= 4049)) then

				return true
			end
		end
	end

	return false
end

 
function IsHom (id) 
------------------------------------------
-- '_'_z___"_N_<_X'c">'f (True:_z__ False:'>''_)
-- _z__'id < 100000 '''__''_B'+''_v___C_"_['_z__'">_B
------------------------------------------
	if (id < 100000 and IsMonster(id) ~= 1) then					-- '_mob'c''_''_B
		local type = GetV(V_HOMUNTYPE,id)					-- id'_^_C_v''_''_B

		if (type == nil) then	--     
			type = MobID[id]
		end

		if (type ~= nil and 1 <= type and type <= 16) then
			return true
		end
	end

	return false
end

 
function IsMerc (id) 
	if (id < 100000 and IsMonster(id) ~= 1) then					-- '_mob'c''_''_B
		local type = GetV(V_HOMUNTYPE,id)					-- id'_^_C_v''_''_B

		if (type == nil) then	--     
			type = MobID[id]
		end

		if (type ~= nil and 17 <= type and type <= 46) then
			return true
		end
	end

	return false
end


function IsPet (id) 
------------------------------------------
-- Ώۂybgf (True:ybg False:̑)
--
-- ybǵAV_HOMUNTYPE̖߂l͕ʂMobƓAIsMonster  0 Ȃ̂Ń\`FbNB
-- (mob^NPCɂȗ\B)
------------------------------------------

	if (IsMonster(id) ~= 1) then					-- '_mob'c''_''_B
		local type = GetV(V_HOMUNTYPE,id)					-- id̃^Cv𒲂ׂB

		if (type == nil) then	--     
			type = MobID[id]
		end

		if (type ~= nil and 1000 <= type and type <= 1700) then				-- G̏ꍇType1001`1700 Ȃ烂X^[FgpID܂܂邯ǋCɂȂB
			return true
		end
	end

	return false
end

 
function IsPlant(id) 
------------------------------------------
-- ARMOB`FbN
------------------------------------------

-- zɂĂsŖRMob͊܂߂ĂȂI

	local type = GetV(V_HOMUNTYPE,id)					-- id̃^Cv𒲂ׂB

	if (type == nil) then	--     
		type = MobID[id]
	end

	if (type == nil) then
		return false
	elseif (1078 <= type and type <= 1085) then
		return true										-- ^[Qbgor̂̏ꍇ
--	elseif (type == 1008 or type == 1047 or type == 1048 or type == 1097 or type == 1142) then
--		return true										-- ^[QbgRMOB̏ꍇ
	elseif (targer == 1589 or type == 1575 or type == 1579 or type == 1555 or type == 1590) then
		return true										-- ^[QbgoCIvg̏ꍇ
	end

	return false
end
